#include "main.h"

#define F_CPU 48000000UL
#define CYCLES_IN_DLYTICKS_FUNC     8
#define MS_TO_DLYTICKS(ms)          (U32)(F_CPU / 1000 * ms / CYCLES_IN_DLYTICKS_FUNC) // ((float)(F_CPU)) / 1000.0
#define DelayTicks(ticks)           {volatile U32 n=ticks; while(n--);}//takes 8 cycles
#define DelayMs(ms)                 DelayTicks(MS_TO_DLYTICKS(ms))//uses 20bytes

struct usart_module usart_instances[UDI_CDC_PORT_NB];

struct usart_config usart_cfgs[UDI_CDC_PORT_NB];

int cdc_cfg_seq[UDI_CDC_PORT_NB] = {0, 0};

static volatile uint8_t main_b_cdc_enable = 0;
static volatile uint8_t main_port_dtr = 0;
static volatile uint8_t main_port_rts = 0;

// CDC data
static uint8_t tx_data[UDI_CDC_PORT_NB];
static uint8_t rx_data[UDI_CDC_PORT_NB];

static volatile uint8_t tx_callback_flag[UDI_CDC_PORT_NB] = {0, 0};

enum usart_character_size lookup_charlength(uint8_t bits) {
	switch (bits) {
		case 5:     return USART_CHARACTER_SIZE_5BIT;
		case 6:     return USART_CHARACTER_SIZE_6BIT;
		case 7:     return USART_CHARACTER_SIZE_7BIT;
		default:
		case 8:     return USART_CHARACTER_SIZE_8BIT;
		case 9:     return USART_CHARACTER_SIZE_9BIT;
	}
}

enum usart_stopbits lookup_stopbits(uint8_t stopbits) {
	switch (stopbits) {
		case CDC_STOP_BITS_2:     return USART_STOPBITS_2;
		default:
		case CDC_STOP_BITS_1:     return USART_STOPBITS_1;
	}
}

enum usart_parity lookup_paritytype(uint8_t parity) {
	switch (parity) {
		case CDC_PAR_EVEN:  return USART_PARITY_EVEN;
		case CDC_PAR_ODD:   return USART_PARITY_ODD;
		default:
		case CDC_PAR_NONE:  return USART_PARITY_NONE;
	}
}

static void usart_tx_callback(struct usart_module *const module) {
	uint8_t port = module->hw == SERCOM1 ? 0 : 1;

	tx_callback_flag[port] = 1;
	/* Data ready to be sent */
	if (udi_cdc_multi_is_rx_ready(port)) {
		/* Transmit next data */
		tx_data[port] = udi_cdc_multi_getc(port);
		usart_write_buffer_job(module, &tx_data[port], 1);
	} else {
		/* Fifo empty then Stop UART transmission */
		usart_disable_callback(module, USART_CALLBACK_BUFFER_TRANSMITTED);
	}
	tx_callback_flag[port] = 0;
}

static void usart_rx_callback(struct usart_module *const module) {
	/* Data received */
	uint8_t port = module->hw == SERCOM1 ? 0 : 1;

	/* Transfer UART RX fifo to CDC TX */
	if (!udi_cdc_multi_is_tx_ready(port)) {
		/* Fifo full */
	} else {
		udi_cdc_multi_putc(port, rx_data[port]);
	}

	usart_read_buffer_job(module, &rx_data[port], 1);
}

int main(void) {
	irq_initialize_vectors();
	cpu_irq_enable();

	system_interrupt_enable_global();
	
	sleepmgr_init();

	system_init();
	board_init();

	usart_start();

	udc_start();

	while (true) {
		sleepmgr_enter_sleep();
	}
}

void print_dec(uint8_t port, uint32_t val) {
	int val_dec[10];
	
	if (val == 0) {
		udi_cdc_multi_putc(port, '0');
		return;
	}
	
	int i = 0;
	while (val > 0 && i < 10) {
		val_dec[i] = val % 10;
		val /= 10;
		i++;
	}
	
	for (int k = i-1; k >= 0; k--) {
		udi_cdc_multi_putc(port, '0'+val_dec[k]);
	}
}

void usart_start(void) {
	usb_cdc_line_coding_t cdc_cfg_default = {
		UDI_CDC_DEFAULT_RATE,
		CDC_STOP_BITS_1,
		CDC_PAR_NONE,
		8
	};
	
	for (uint8_t port = 0; port < UDI_CDC_PORT_NB; port++) {
		main_cdc_set_coding(port, &cdc_cfg_default);
		main_cdc_set_coding(port, &cdc_cfg_default);
	}
}

void main_tx_empty_notify(uint8_t port) {
	
}

void main_rx_notify(uint8_t port) {
	if (!tx_callback_flag[port]) {
		/* Transmit first data */
		usart_enable_callback(&usart_instances[port], USART_CALLBACK_BUFFER_TRANSMITTED);
		tx_data[port] = udi_cdc_multi_getc(port);
		usart_write_buffer_job(&usart_instances[port], &tx_data[port], 1);
	}
}

void main_cdc_set_coding(uint8_t port, const usb_cdc_line_coding_t* cfg) {
	struct usart_module* module = &usart_instances[port];
	struct usart_config* module_cfg = &usart_cfgs[port];
	
	cdc_cfg_seq[port]++;
	
	if (cdc_cfg_seq[port] == 1) {
		usart_get_config_defaults(module_cfg);
		
		// first line encoding message, only baudrate is valid (WTF?)
		module_cfg->baudrate = cfg->dwDTERate;
		return;
	}
	
	// second line encoding message, baudrate is wrong, everything else OK
	module_cfg->stopbits       = lookup_stopbits(cfg->bCharFormat);
	module_cfg->parity         = lookup_paritytype(cfg->bParityType);
	module_cfg->character_size = lookup_charlength(cfg->bDataBits);

	cdc_cfg_seq[port] = 0;
	
	module->receiver_enabled = false;
	module->transmitter_enabled = false;
	
	if (port == 0) {
		module_cfg->mux_setting = USART_RX_3_TX_2_XCK_3;
		module_cfg->pinmux_pad0 = PINMUX_UNUSED;
		module_cfg->pinmux_pad1 = PINMUX_UNUSED;
		module_cfg->pinmux_pad2 = PINMUX_PA08C_SERCOM1_PAD2;
		module_cfg->pinmux_pad3 = PINMUX_PA09C_SERCOM1_PAD3;
	} else {
		module_cfg->mux_setting = USART_RX_1_TX_0_XCK_1;
		module_cfg->pinmux_pad0 = PINMUX_PA14C_SERCOM0_PAD0;
		module_cfg->pinmux_pad1 = PINMUX_PA15C_SERCOM0_PAD1;
		module_cfg->pinmux_pad2 = PINMUX_UNUSED;
		module_cfg->pinmux_pad3 = PINMUX_UNUSED;
	}
	module_cfg->start_frame_detection_enable = true;

	usart_disable_callback(module, USART_CALLBACK_BUFFER_RECEIVED);
	usart_disable_callback(module, USART_CALLBACK_BUFFER_TRANSMITTED);
	
	usart_reset(module);
	
	while (usart_init(module, port == 0 ? SERCOM1 : SERCOM0, module_cfg) != STATUS_OK) {

	}
	
	usart_enable(module);
	
	usart_register_callback(module, (usart_callback_t)usart_rx_callback, USART_CALLBACK_BUFFER_RECEIVED);
	usart_register_callback(module, (usart_callback_t)usart_tx_callback, USART_CALLBACK_BUFFER_TRANSMITTED);
	usart_enable_callback(module, USART_CALLBACK_BUFFER_RECEIVED);
	usart_enable_callback(module, USART_CALLBACK_BUFFER_TRANSMITTED);
	
	// start reading
	usart_read_buffer_job(module, &rx_data[port], 1);
}

void main_suspend_action(void) {
	//ui_powerdown();
}

void main_resume_action(void) {
	//ui_wakeup();
}

void main_sof_action(void) {
	//if (!main_b_cdc_enable)
	//	return;
	//ui_process(udd_get_frame_number());
}

bool main_cdc_enable(uint8_t port) {
	main_b_cdc_enable |= (1 << port);
	return true;
}

void main_cdc_disable(uint8_t port) {
	main_b_cdc_enable &= ~(1 << port);
	//main_b_cdc_enable = false;
	main_port_rts = 0;
	main_port_dtr = 0;
}

void main_cdc_set_dtr(uint8_t port, bool b_enable) {
	if (b_enable) {
		main_port_dtr |= (1 << port);
	} else {
		main_port_dtr &= ~(1 << port);
	}
}

void main_cdc_set_rts(uint8_t port, bool b_enable) {
	if (b_enable) {
		main_port_rts |= (1 << port);
		} else {
		main_port_rts &= ~(1 << port);
	}
}
